#!/usr/stsci/pyssg/Python-2.5.4/bin/python
#
#NAME:
#  SatelliteControl.py 
#
#PURPOSE:
#  To simulate the control system for the CXBN Nanosat.  This involves 
#  propogating the orbit of the center of mass of the system and 
#  modeling the rotational dynamics of the system. 
#
#KEYWORDS:
#  DateTime: str
#    The date and time for which we want to calculate the observation.  This should be 
#    in the pyephem prefered format of "Year/Month/Day Hour:Minute:Second" in UT.
#    The date and time for which we want to calculate the observation.  This should be 
#  TLE_File: str
#    The path to the TLE file that will be used as input for creating the PyEphem 
#    satellite object
#  ESolveMethod: int
#    0 - Solve Euler Equations using a first order Euler Method
#    1 - Solve Euler Equations using a fourth order Runge-Kutta method
#  TestGravityBoom: int
#    0 - Perform simulation with all input torques and controls
#    1 - Perform simulation of a gravity gradient stabilized satellite to 
#        demonstrate that the dynamics of the simulation are correct.  
#        For this case, the moment of inertia tensor will be highly concentrated 
#        along the z-axis. 
#  UseGGTorque: int
#    0 - Don't calculate and apply the gravitational gradient torque
#    1 - Calculate and apply the gravitational gradient torque
#  UseAeroTorque: int
#    0 - Don't calculate and apply the aerodynamic torque
#    1 - Calculate and apply the aerodynamic torque
#  UseMagTorque: int
#    0 - Don't calculate and apply the magnetic torque due to the magnetorquers on the CubeSat
#    1 - Calculate and apply the magnetic torque due to the magnetorquers on the CubeSat
#  TrackSun: int
#    0 - Don't track the sun's location
#    1 - Track the sun's location and attempt to steer the satellite toward it
#  WriteBField: int
#    0 - Dont' write out an ASCII file with the B field
#    1 - Write an ASCII file with the B field
#
#NOTES:
#  The pyephem package actually checks the CheckSum on each line of the TLE file.  One must make 
#  sure that the last digit of the line equals the sum of all digits (modulo 10).  One easy way 
#  to make this true is just to play with the last digit until it works.
#
#  There are three reference coordinate systems used here:
#    e = Earth Centered Inertial Coordinate System
#    r = Satellite Reference Frame (-z points toward center of earth, +x points in direction of velocity)
#    b = Body Coordinate System (axes point along principal axes)
#
#EXAMPLES:
#  GRAVITY BOOM-----
#  run SatelliteControl.py --TestGravityBoom=1 --TLE_File='./TLE/GravityGradientTest.tle'
#  MAGNETIC CONTROL DETUMBLE------
#  run SatelliteControl.py --UseMagTorque=1 --UseGGTorque=0 --UseAeroTorque=0 --TrackSun=0
#
###############################################################################################################

import os
execfile('./CXBNNanoSatSetup.py')

import ephem
from datetime import *
from CXBN_Utilities import *

# Main routine if we are called as a script
if __name__ == "__main__":

  ##PARSE KEYWORDS
  keywords    = ['DateTime=', 'TLE_File=', 'ESolveMethod=', 'TestGravityBoom=', \
                 'UseGGTorque=', 'UseAeroTorque=', 'UseMagTorque=', 'TrackSun=', \
                 'WriteBField='] 

  #DEFAULT VALUES FOR KEYWORDS
  DateTime        = '2012/01/21 20:20:00'
  TLE_File        = './TLE/CXBNTLE.txt'
  ESolveMethod    = 1
  TestGravityBoom = 0
  UseGGTorque     = 1
  UseAeroTorque   = 1
  UseMagTorque    = 0
  TrackSun        = 0
  WriteBField     = 0

  #NOW PARSE KEYWORDS
  opts, extraparams = getopt.getopt(sys.argv[1:],'',keywords)
  for o,p in opts:
    if o in ['--DateTime']:
      DateTime = p
    elif o in ['--TLE_File']:
      TLE_File = p
    elif o in ['--ESolveMethod']:
      ESolveMethod = int(p)
    elif o in ['--TestGravityBoom']:
      TestGravityBoom = int(p)
    elif o in ['--UseGGTorque']:
      UseGGTorque = int(p)
    elif o in ['--UseAeroTorque']:
      UseAeroTorque = int(p)
    elif o in ['--UseMagTorque']:
      UseMagTorque = int(p)
    elif o in ['--TrackSun']:
      TrackSun = int(p)
    elif o in ['--WriteBField']:
      WriteBField = int(p)

  #OUTPUT FILES 
  if WriteBField == 1:
    if not os.path.exists('./Data/') : os.mkdir('./Data/')
    BField_FH = open('./Data/BFieldVsTime.txt', 'w')

  #############################################################################################################
  ################################ DATE AND TIME ##############################################################
  #Format the time appropriately
  Year  = int(DateTime[0:4])
  Month = int(DateTime[5:7])
  Day   = int(DateTime[8:10])
  Hours = int(DateTime[11:13])
  Mins  = int(DateTime[14:16])
  Secs  = int(DateTime[17:19])
  DateStr  = DateTime[0:10]
  DateTime = datetime(Year, Month, Day, Hours, Mins, Secs)

  #Convert this into a PyEphem Date (# days since noon on 1899 December 31) 
  EpDate_0 = ephem.date(DateTime)
  EpDate   = EpDate_0

  #############################################################################################################
  ################################ SATELITTE AND OTHER OBJECTS ################################################
  #Create an instance of the constant object
  c = c()

  #Call getsat to return an ephemeris satellite object based on the TLE
  CXBN  = getsat(TLE_File, EpDate)
  CXBN2 = getsat(TLE_File, EpDate)

  #Create a satellite object
  cxbn = CubeSat()

  #Create an oject for the sun if we're trying to point to it
  if (TrackSun == 1):
    Sun = ephem.Sun()
    Sun.compute(EpDate)
    t_sun_interval = 50
    r_sun_eci      = vec3()
    r_sun_b        = vec3()
 
  #############################################################################################################
  ############################### INTIIAL POSITION AND VELOCITY ###############################################
  #Get the initial position and velocity vectors of the satellite to define the 
  #reference and body coordinate systems
  dt_sub = 0.00001
  CXBN.compute(EpDate-dt_sub*ephem.second) ; r_eci_t_minus_1 = getSatECI(CXBN,c)
  CXBN.compute(EpDate+dt_sub*ephem.second) ; r_eci_t_plus_1  = getSatECI(CXBN,c)
  CXBN.compute(EpDate)                     ; r_eci_t         = getSatECI(CXBN,c)
  v_t = (1/(2*dt_sub))*(r_eci_t_plus_1-r_eci_t_minus_1)
  
  #Now normalize these and find the cross product 
  v_t_norm = v_t.normalize()
  r_eci_norm = r_eci_t.normalize()
  r_v_cross_norm = r_eci_norm.cross(v_t_norm).normalize()
  
  #Get the initial position vector of the satellite
  CXBN.compute(EpDate)
  r_eci      = getSatECI(CXBN, c)
  r_eci_prev = r_eci

  #############################################################################################################
  ############################### TIME VECTORS ################################################################
  #Now calculate the approximate velocity vector at this time-step
  EpDate_t_minus_1 = EpDate

  #Set the time increment for the propogation
  dt       = double(0.01)          #Time difference in seconds
  t_final  = 1*600.               #Final time in seconds
  NumTimes = t_final/dt
  Times    = arange(NumTimes)

  #Set the initial time to t=0
  Time=0

  #############################################################################################################
  ############################### TRACKING ARRAYS  ############################################################
  #Arrays for tracking variables (not necessary for functioning)
  w_x_Arr  = zeros(NumTimes, dtype=float32)
  w_y_Arr  = zeros(NumTimes, dtype=float32)
  w_z_Arr  = zeros(NumTimes, dtype=float32)
  t_x_Arr  = zeros(NumTimes, dtype=float32)
  t_y_Arr  = zeros(NumTimes, dtype=float32)
  t_z_Arr  = zeros(NumTimes, dtype=float32)
  #q_0_Arr  = zeros(NumTimes, dtype=float32)
  #q_1_Arr  = zeros(NumTimes, dtype=float32)
  #q_2_Arr  = zeros(NumTimes, dtype=float32)
  #q_3_Arr  = zeros(NumTimes, dtype=float32)
  #EleArr   = zeros(NumTimes, dtype=float32)
  #VelArr   = zeros(NumTimes, dtype=float32)
  #GTorArr  = zeros(NumTimes, dtype=float32)
  #ATorArr  = zeros(NumTimes, dtype=double)
  #w_t_Arr  = zeros(NumTimes, dtype=float32)
  #L_t_Arr  = zeros(NumTimes, dtype=float32)
  B_Ang_Arr = zeros(NumTimes, dtype=float32)
  S_Ang_Arr = zeros(NumTimes, dtype=float32)

  #############################################################################################################
  ############################### BASIS VECTORS AND COORDINATE SYSTEMS#########################################
  #Set up the basis vectors for the inertial frame
  i   = vec3(1,0,0)   ; j   = vec3(0,1,0)    ; k   = vec3(0,0,1)
  e_i = vec3(1,0,0)   ; e_j = vec3(0,1,0)    ; e_k = vec3(0,0,1)
  r_i = r_eci_norm    ; r_j = r_v_cross_norm ; r_k = v_t_norm
  b_i = r_eci_norm    ; b_j = r_v_cross_norm ; b_k = v_t_norm

  #Get the initial Direction Cosine Matrices
  e_2_b=ReturnDCM(e_i, e_j, e_k, b_i, b_j, b_k)
  r_2_b=ReturnDCM(r_i, r_j, r_k, b_i, b_j, b_k)

  #############################################################################################################
  ############################### ANGULAR VELOCITIES AND QUATERNIONS ##########################################
  #Set the initial angular velocity
  #Expected rates for detumbling are 20 degrees/second
  omega    = vec3()     #Set up an angular velocity vector)
  if (UseMagTorque == 1):
    omega.x  = 18.5*pi/180                  #x component of angular velocity (w_1 = p)
    omega.y  = -25.0*pi/180                 #y component of angular velocity (w_2 = q)
    omega.z  = 12.2*pi/180                  #z component of angular velocity (w_3 = r)
  elif (TrackSun == 1):
    omega.x  = 2.85*pi/180.0                #x component of angular velocity (w_1 = p)
    omega.y  = -1.20*pi/180.0               #y component of angular velocity (w_2 = q)
    omega.z  = 3.00*pi/180.0                #z component of angular velocity (w_3 = r)

  else:
    omega.x  = 0.800                        #x component of angular velocity (w_1 = p)
    omega.y  = 0.400                        #y component of angular velocity (w_2 = q)
    omega.z  = -.900                        #z component of angular velocity (w_3 = r)
  
  #Now copy the angular velocity for other uses
  omega_0  = omega                          #This is the initial omega, does not change
  omega_t0 = omega                          #This is omega at timestep t[i]  
  omega_t1 = omega                          #This is omega at timestep t[i+1]

  #Get the initial quaternion
  quat_initial = quat(e_2_b)                #Get the DCM from the quaternion
  quat_initial = quat_initial.normalize()   #normalize the quaternion
  q_t0 = vec4()                             #quaternion at time step [t]
  q_t1 = vec4()                             #quaternion at time step [t+1]
  q_t0[3] = quat_initial.w                  
  q_t0[0] = quat_initial.x
  q_t0[1] = quat_initial.y
  q_t0[2] = quat_initial.z
  q_t1 = q_t0

  #############################################################################################################
  ############################### TORQUES AND MAGNETIC FIELD VECTORS ##########################################

  #Set up initial torque vectors
  aero_torque = vec3(0,0,0)                 #The 3 vector for the aerodynamic torque
  gg_torque   = vec3(0,0,0)                 #The 3 vector for the gravity gradient torque
  mag_torque  = vec3(0,0,0)                 #The 3 vector for the magnetic torque 

  #Set up arrays for the magnetic fields
  B_Earth_b   = vec3(0,0,0)                 #Earth's magnetic field in the body frame of the satellite
  k_t_loop    = 0.001                       #The proportional factor for the control loop
  k_w_loop    = -0.10                       #The derivative factor for the control looop
  m_mag       = 0.166                       #The magnetic moment magnitude in A*m^2 
  omega_des   = vec3(0,0,0)                 #The desired angular veloctiy in rad/s 

  #Keep track of the r_eci to do magnetic field calculations only every so often
  min_dist_for_mag_field = 320000           #Minimum distance to calculate new B field (in meters)
  del_t_mag_field    = 0.5                  #The time interval for computing the magnetic field (in seconds)
  SenseTorqueFreq    = 1.                   #Frequency at which sensing/torquing takes place
  SenseTorquePeriod  = 1/SenseTorqueFreq    #Period of sensing/torquing cycle
  mag_duty_cycle     = 0.5                  #Fraction of each period spent measuring the magnetic field 
  r_eci_mag_prev     = r_eci                #Update the coordinates of the previous magnetic field reading
  m_max              = 0.1666               #Maximum magnetic moment that can be applied along one axis 
  K_B                = 20.0000              #Proportional factor for the damping torque
  ApplyMagFieldNow   = 0                    #Boolean for applying the magnetic field

  #Set up the magnetic field vectors to hold the magnetic field of current and previous step in body coords 
  B_Earth_b_tm0 = vec3()
  B_Earth_b_tm1 = vec3()

  #Get the magnetic field of earth in eci coordinates
  B_Earth0_eci_v   = getmagfield(CXBN, DateStr)
  t_prev_mag_field = Time

  #Inrement the time for the next reading and recompute the magnetic field for that time
  EpDate2 = double(EpDate) + ephem.second*del_t_mag_field
  CXBN2.compute(EpDate2)
  B_Earth1_eci_v = getmagfield(CXBN2, DateStr)

  #Compute the difference in magnetic field from r_eci_previous
  del_B_Earth_eci   = vec3()
  del_B_Earth_eci.x = B_Earth1_eci_v.x-B_Earth0_eci_v.x
  del_B_Earth_eci.y = B_Earth1_eci_v.y-B_Earth0_eci_v.y
  del_B_Earth_eci.z = B_Earth1_eci_v.z-B_Earth0_eci_v.z
 
  #And get the next time the magnetic field should be updated
  t_next_mag_field = Time + del_t_mag_field
 
  #Get the initial date and orientation
  CXBN.compute(EpDate)
  r_eci = getSatECI(CXBN, c)

  #Allow for a test of the gravity boom
  if TestGravityBoom == 1:
    execfile('./Test/GravityBoomConfigurationForSatelliteControl.py')

  #Reset the time in case we did some manipulation
  EpDate = EpDate_0
  Time   = 0

  #############################################################################################################
  #PROPAGATE THE SATELLITE IN ITS ORBIT
  #############################################################################################################
  for t_i in Times:  

    #Inrement the time 
    EpDate = double(EpDate) + ephem.second*dt
    Time  += dt

    #Recompute the solution with the SGP4
    CXBN.compute(EpDate)

    #Get the position of the satellite in Earth-Centered Inertial Coordinates
    r_eci = getSatECI(CXBN, c) 
    v_eci = (1/dt)*(r_eci - r_eci_prev)
    r_eci_prev = r_eci

    #Now normalize these, find the cross product, and assign to the reference vectors 
    r_i = v_eci.normalize()
    r_k = r_eci.normalize()
    r_j = r_k.cross(r_i).normalize()

    #Now get the Torques, gg and aero
    if UseGGTorque == 1: 
      gg_torque = getGravityGradientTorque(r_eci, e_2_b, cxbn, c)
    if UseAeroTorque == 1:
      aero_torque = getAerodynamicTorque(v_eci, r_eci, e_2_b, cxbn, c)

    #Update the torque using the gravity gradient and aerodynamic.  Magnetic is below
    tot_torque  = gg_torque+aero_torque

    #Calculate the magnetic field if it's necessary (i.e. after a certain time interval)
    if (Time >= t_next_mag_field):

      #Get the magnetic field of earth in eci coordinates
      B_Earth0_eci_v = getmagfield(CXBN, DateStr)
      t_prev_mag_field = Time

      #Inrement the time for the next reading and recompute the magnetic field for that time
      EpDate2 = double(EpDate) + ephem.second*del_t_mag_field
      CXBN2.compute(EpDate2)
      B_Earth1_eci_v = getmagfield(CXBN2, DateStr)

      #Compute the difference in magnetic field from r_eci_previous
      del_B_Earth_eci   = vec3()
      del_B_Earth_eci.x = B_Earth1_eci_v.x-B_Earth0_eci_v.x
      del_B_Earth_eci.y = B_Earth1_eci_v.y-B_Earth0_eci_v.y
      del_B_Earth_eci.z = B_Earth1_eci_v.z-B_Earth0_eci_v.z

      #And get the time difference and exit the loop since we've got the difference
      t_next_mag_field = Time + del_t_mag_field

    #Update the magnetic field in the body of the spacecraft
    B_Earth_eci_v  = B_Earth0_eci_v + ((Time-t_prev_mag_field)/(del_t_mag_field))*del_B_Earth_eci
    B_Earth_b_tm0  = e_2_b * B_Earth_eci_v

    #Determine where the sun is in the body frame of the spacecraft
    if (TrackSun == 1) :#and (int(dt*t_i) % t_sun_interval == 0)):
      r_sun_eci.x = sin(Sun.dec)*cos(Sun.ra)      
      r_sun_eci.y = sin(Sun.dec)*sin(Sun.ra)
      r_sun_eci.z = cos(Sun.dec)
      r_sun_b     = e_2_b*r_sun_eci

      #Get the solar panel illumination for each quadrant
      SolarIlluminationArr = getSolarPanelIllumination(r_sun_b, cxbn)

      #Now try to orient toward the sun
      csun_torque = ReturnCoarseSunTorque(SolarIlluminationArr, B_Earth_b_tm0, omega_t0, \
                                          k_t_loop, k_w_loop, m_mag)
      tot_torque = tot_torque+csun_torque

      #Get the angle between the z-axis and the sun pointing vector
      S_Ang_Arr[t_i] = math.acos(k.normalize()*r_sun_b.normalize())


    #Determine the magnetic field and apply an appropriate torque
    if (UseMagTorque == 1):

        #See where in the 60 Hz cycle we are
        DecimalTime2 = float('%.2f' % Time) 
        NumberOfPeriods = int(DecimalTime2/SenseTorquePeriod) 
        Phase = DecimalTime2/SenseTorquePeriod - NumberOfPeriods

        if (Phase < mag_duty_cycle):
          #Update the current magnetic field reading (at t-minus-0) 
          B_Earth_eci_v  = B_Earth0_eci_v + ((Time-t_prev_mag_field)/(del_t_mag_field))*del_B_Earth_eci 
          B_Earth_b_tm0  = e_2_b * B_Earth_eci_v 

        elif (Phase == mag_duty_cycle):
    
          #Update the magnetic field at the current step 
          mag_moment, mag_torque = ReturnMagFieldAlignTorque(B_Earth_b_tm0, B_Earth_b_tm1, K_B, m_max, dt)

          #Now update the previous magnetic field readings (at t-minus-1 and t-minus-2)
          B_Earth_b_tm1  = B_Earth_b_tm0
          ApplyMagFieldNow = 1

        #Apply a control torque
        elif ((Phase > mag_duty_cycle) and (ApplyMagFieldNow ==1)):
          B_Earth_eci_v  = B_Earth0_eci_v + ((Time-t_prev_mag_field)/(del_t_mag_field))*del_B_Earth_eci
          B_Earth_b_cur  = e_2_b * B_Earth_eci_v
          mag_torque = mag_moment.cross(B_Earth_b_cur)
          tot_torque = tot_torque+mag_torque  

        #Calculate the angle between the z-axis and the magnetic field
        B_Ang_Arr[t_i] = math.acos(k.normalize()*B_Earth_b_tm0.normalize()) 

    #Update the angular velocity by Solving Euler's equations from [t] to [t+1]
    omega_t1 = SolveEuler(omega_t0, tot_torque, cxbn.InertiaTensor, dt, ESolveMethod)
    w_x_Arr[t_i] = omega_t1.x
    w_y_Arr[t_i] = omega_t1.y
    w_z_Arr[t_i] = omega_t1.z
    #t_x_Arr[t_i] = mag_torque.x
    #t_y_Arr[t_i] = mag_torque.y
    #t_z_Arr[t_i] = mag_torque.z

    #Now get the quaternion at time [t+1]
    q_t1, q_t1_quat = UpdateQuaternion(q_t0, dt, omega_t0, omega_t1)

    #Get the updated earth to body rotation matrix and the basis vectors
    e_2_b = q_t1_quat.toMat3()
    b_2_e = e_2_b.transpose()
    b_i = b_2_e*i
    b_j = b_2_e*j
    b_k = b_2_e*k

    #Now get the Direction Cosine Matrix between the reference and body frames
    r_2_b=ReturnDCM(r_i, r_j, r_k, b_i, b_j, b_k)
    
    #And convert these to Euler Angles
    Yaw   = math.atan2(r_2_b[0,1], r_2_b[0,0])
    Pitch = -math.asin(r_2_b[0,2])
    Roll  = math.atan2(r_2_b[1,2], r_2_b[2,2]) 

    #Update omega(t+1) = omega(t) for the next time step
    omega_t0 = omega_t1
    q_t0     = q_t1

    #If we need to write the magnetic field, do so
    if WriteBField == 1:
      B_x_str = "%.10f" % B_Earth_b_tm0.x
      B_y_str = "%.10f" % B_Earth_b_tm0.y
      B_z_str = "%.10f" % B_Earth_b_tm0.z
      BField_FH.write(str(Time)+'\t\t'+B_x_str+'\t\t'+B_y_str+'\t\t'+B_z_str+'\n')
     
##############################################################################
#Now plot the angular velocities
mplot.figure(0)
mplot.clf()
mplot.plot(Times*dt, w_x_Arr, label=r'$\omega_{x}$')
mplot.plot(Times*dt, w_y_Arr, label=r'$\omega_{y}$')
mplot.plot(Times*dt, w_z_Arr, label=r'$\omega_{z}$')
mplot.xlabel('Time (seconds)')
mplot.ylabel('Angular Velocity (radians/sec)')
mplot.legend()

if (UseMagTorque):
  mplot.figure(1)
  mplot.clf()
  mplot.plot(Times*dt, B_Ang_Arr)
  mplot.xlabel('Time (seconds)')
  mplot.ylabel('Angle between z-axis and Magnetic Field Vector')

#Option to plot the separation from the sun pointing vector
if (TrackSun == 1):
  mplot.figure(1)
  mplot.clf()
  mplot.plot(Times*dt, S_Ang_Arr)
  mplot.xlabel('Time (seconds)')
  mplot.ylabel('Angle between z-axis and Sun Vector')

#Close the BField File if necessary
if WriteBField == 1:
  BField_FH.close()
